# -*- coding: utf8 -*-
from __future__ import unicode_literals

import logging

from driverng import Worklist
from driverng.astm import (
    AstmMessageV2,
    HeaderSegment,
    HprimDate,
    HprimForeign,
    HprimList,
    HprimMultiple,
    HprimString,
    Segment,
)

logger = logging.getLogger("driverng.dxi9000")


class Dxi9000SegmentH(HeaderSegment):
    sender_name = HprimString((5, 1))
    sender_id = HprimString((5, 2))
    software_version = HprimString((5, 3))
    version = HprimString(13)
    date = HprimDate(14)

    patients = HprimForeign("P", "header")
    queries = HprimForeign("Q", "header")
    end = HprimForeign("L", "header", one2one=True, required=True)
    infos = HprimForeign("M", "header")
    orders = HprimForeign("O", "header")


class Dxi9000SegmentQ(Segment):
    specimen_id = HprimString((3, 2))
    status_code = HprimString(13)


class Dxi9000SegmentP(Segment):
    patient_id = HprimString(3)
    laboratory_patient_id = HprimString(4)
    last_name = HprimString((6, 1))
    first_name = HprimString((6, 2))
    middle_name = HprimString((6, 3))
    birthdate = HprimDate(8, fmt_date="%Y%m%d")
    sex = HprimString(9)
    attending_id = HprimString((14, 1))
    attending_last_name = HprimString((14, 2))
    attending_first_name = HprimString((14, 3))
    attending_middle_name = HprimString((14, 4))
    attending_prefix = HprimString((14, 6))
    patient_location = HprimString(26)

    orders = HprimForeign("O", "patient")


class Dxi9000SegmentO(Segment):
    specimen_id = HprimString(3)
    tests = HprimList(
        5,
        basefield=HprimMultiple(
            0,
            test_id=HprimString(1),
            name=HprimString(2),
            test_id_type=HprimString(3),
            local_code=HprimString(4),
        ),
    )
    priority = HprimString(6)
    collection_date = HprimDate(8)
    action_code = HprimString(12)
    sample_type = HprimString(16)
    prescriber_code = HprimString((17, 1))
    prescriber_last_name = HprimString((17, 2))
    prescriber_first_name = HprimString((17, 3))
    prescriber_title = HprimString((17, 6))
    report_type = HprimString(26)

    results = HprimForeign("R", "order")
    comments = HprimForeign("C", "order")


class Dxi9000SegmentR(Segment):
    test_id = HprimString((3, 1))
    test_name = HprimString((3, 2))
    test_type = HprimString((3, 3))
    local_code = HprimString((3, 4))
    primary_result = HprimString((4, 1))
    value = HprimString((4, 2))
    interpretation = HprimString((4, 3))
    units = HprimString(5)
    reference_range = HprimString(6)
    result_status = HprimString(9)
    date_completed = HprimDate(13)
    serial_number = HprimString((14, 1))
    analyzer_model = HprimString((14, 2))
    manufacturer = HprimString((14, 3))

    infos = HprimForeign("M", "result")


class Dxi9000SegmentC(Segment):
    comment_source = HprimString(3)
    comment_id = HprimString((4, 1))
    comment_description = HprimString((4, 2))
    comment_type = HprimString(5)


class Dxi9000SegmentM(Segment):
    record_type_id = HprimString(3)
    identifier_id = HprimString((4, 1))
    identifier_name = HprimString((4, 2))
    status_id = HprimString((5, 1))
    status_text = HprimString((5, 2))
    type_id = HprimString((6, 1))
    type_text = HprimString((6, 2))
    container_identifier = HprimString(7)
    lot_number = HprimString(8)


class Dxi9000SegmentL(Segment):
    terminaison_code = HprimString(3, default="N")


class Driver(AstmMessageV2):
    SPEC = {"code": "dxi9000", "label": "DxI 9000"}

    SEGMENT_CLASS_MAP = {
        "H": Dxi9000SegmentH,
        "P": Dxi9000SegmentP,
        "O": Dxi9000SegmentO,
        "Q": Dxi9000SegmentQ,
        "R": Dxi9000SegmentR,
        "M": Dxi9000SegmentM,
        "C": Dxi9000SegmentC,
        "L": Dxi9000SegmentL,
    }
    DEFAULT_SEPARATORS = "|\\^&"

    def create(self, **kwargs):
        message = Dxi9000SegmentH(type_segment="H", **kwargs)
        message.end = Dxi9000SegmentL(type_segment="L")
        return message

    def encode_worklist(self, worklist, context):
        if worklist.context != "order":
            raise NotImplementedError("context %s" % (worklist.context,))
        message = self.create()
        for patient in worklist.get_astm_worklist(
            unique_together=["action", "priority"],
            priority={"A": "S", "S": "S", worklist.ELSE: "R"},
            action={"del": "C", "add": "A", "new": "N"},
        ):
            attending = [p for p in patient.physicians if p.type.lower() == "attending"]
            prescriber = [
                p for p in patient.physicians if p.type.lower() == "prescriber"
            ]
            seg_p = message.patients.add(
                patient_id=patient.patient_id,
                laboratory_patient_id=patient.lab_patient_id,
                last_name=patient.last_name,
                first_name=patient.first_name,
                middle_name=patient.maiden_name,
                birthdate=patient.birthdate,
                sex=patient.sex,
                attending_id=attending[0].code if attending else "",
                attending_last_name=attending[0].last_name if attending else "",
                attending_first_name=attending[0].first_name if attending else "",
                attending_prefix=attending[0].title if attending else "",
                patient_location=patient.location,
            )
            if patient.patient_comment1:
                seg_p.commentaires.add(
                    comment_source="L",
                    comment_description=patient.patient_comment1,
                    comment_type="G",
                )
            if patient.patient_comment2:
                seg_p.commentaires.add(
                    comment_source="L",
                    comment_description=patient.patient_comment2,
                    comment_type="G",
                )
            for sample, action, priority in patient.samples:
                seg_o = seg_p.orders.add(
                    specimen_id=sample.sample_id,
                    priority=priority,
                    action_code=action,
                    collection_date=sample.collection_date,
                    sample_type=sample.specimen,
                    prescriber_code=prescriber[0].code if prescriber else "",
                    prescriber_last_name=prescriber[0].last_name if prescriber else "",
                    prescriber_first_name=prescriber[0].first_name
                    if prescriber
                    else "",
                    prescriber_title=prescriber[0].title if prescriber else "",
                    report_type="O",
                )
                for order in sample.orders:
                    t = seg_o.tests.add()
                    t.local_code = order.code
                if sample.request_comment_obr:
                    seg_o.commentaires.add(
                        comment_source="L",
                        comment_description=sample.request_comment_obr.replace(
                            "\r", " "
                        ).replace("\n", " "),
                        comment_type="G",
                    )
        if message.patients:
            return message
        elif context["query"]:
            seg_p = message.patients.add()
            for seg_q in context["query"].queries:
                seg_p.orders.add(specimen_id=seg_q.specimen_id, report_type="Y")
            return message

    def decode_message(self, message):
        if message.queries:
            worklist = Worklist(context="query")
            for seg_q in message.queries:
                worklist.get_or_create(type="patient", sample_id=seg_q.specimen_id)
        else:
            worklist = Worklist(context="result")
            for seg_p in message.patients:
                for seg_o in seg_p.orders:
                    if seg_o.action_code == "Q":
                        sample = worklist.get_or_create(
                            sample_id=seg_o.specimen_id,
                            control_lot=seg_o.specimen_id,
                            type="control",
                        )
                    else:
                        sample = worklist.get_or_create(
                            sample_id=seg_o.specimen_id,
                            specimen=seg_o.sample_type,
                            type="patient",
                        )
                    for seg_r in seg_o.results:
                        comments = []
                        flags = []
                        lots = []
                        if seg_r.commentaires:
                            for seg_c in seg_r.commentaires:
                                comments.append(seg_c.comment_description)
                                if seg_c.comment_type == "I":
                                    flags.append(seg_c.comment_id)
                        if seg_r.infos:
                            for seg_m in seg_r.infos:
                                lots.append(
                                    {
                                        "type": "reagent",
                                        "name": seg_m.identifier_name,
                                        "lot": seg_m.lot_number,
                                    }
                                )
                        sample.results.add(
                            code=seg_r.local_code or seg_o.tests[0].local_code,
                            value=seg_r.primary_result,
                            date=seg_r.date_completed,
                            unit=seg_r.units,
                            instrument="{} {}".format(
                                seg_r.analyzer_model, seg_r.serial_number
                            ),
                            comments=comments,
                            flags=flags,
                            lots=lots,
                        )
                        if seg_r.interpretation:
                            sample.results.add(
                                code="Int_{}".format(seg_r.local_code or seg_o.test_id),
                                value=seg_r.interpretation,
                                date=seg_r.date_completed,
                                instrument="{} {}".format(
                                    seg_r.analyzer_model, seg_r.serial_number
                                ),
                                comments=comments,
                                flags=flags,
                                lots=lots,
                            )

        return worklist
