# -*- coding: utf8 -*-
from __future__ import unicode_literals

import logging
import time

from unidecode import unidecode

from driverng import Worklist
from driverng.astm import (
    AstmMessageV2,
    HeaderSegment,
    Hl7Coding,
    HprimDate,
    HprimForeign,
    HprimList,
    HprimString,
    Segment,
    NoRangSegment,
)

logger = logging.getLogger("driverng.StagoHl7")


class StagoSegmentMSH(HeaderSegment):
    sending_application = HprimString(3)
    sending_facility = HprimString(4)
    receiving_application = HprimString(5, default="sthemE")
    receiving_facility = HprimString(6)
    message_date = HprimDate(7)
    message_type = HprimString((9, 1), default="OML")
    message_code = HprimString((9, 2), default="O33")
    message_label = HprimString((9, 3), default="OML_O33")
    control_id = HprimString(10)
    processing_id = HprimString(11, default="P")
    version_id = HprimString(12, default="2.5.1")
    accept_acknowledgment_type = HprimString(15, default="NE")
    application_acknowledgment = HprimString(16, default="AL")
    character_set = HprimString(18, default="UNICODE UTF-8")
    message_profile = HprimString((21, 1), default="LAB-28")
    message_profile_code = HprimString((21, 2), default="IHE")

    patients = HprimForeign("PID", "header")
    specimens = HprimForeign("SPM", "header")
    infos_samples = HprimForeign("SAC", "patient")
    orders = HprimForeign("OBR", "header")
    queries = HprimForeign("QPD", "header")
    errors = HprimForeign("ERR", "header")
    equals = HprimForeign("EQU", "header")


class StagoSegmentPID(NoRangSegment):
    seq = HprimString(2, default="1")
    patient_id = HprimString(4)
    lastname = HprimString((6, 1))
    firstname = HprimString((6, 2))
    name_type_code = HprimString((6, 7), default="L")
    birthdate = HprimDate(8, fmt_date="%Y%m%d")
    sex = HprimString(9)

    visits = HprimForeign("PV1", "patient")
    orders = HprimForeign("OBR", "patient")
    infos = HprimForeign("NTE", "patient")
    specimens = HprimForeign("SPM", "patient")


class StagoSegmentPV1(Segment):
    seq = HprimString(2, default="1")
    patient_class = HprimString(3, default="N")
    location_room = HprimString((4, 1))
    location_bed = HprimString((4, 2))
    doctor_code = HprimString((8, 1))
    doctor_name = HprimString((8, 2))
    doctor_firstname = HprimString((8, 3))
    service = HprimString((11, 1))

    orders = HprimForeign("OBR", "patient")
    specimens = HprimForeign("SPM", "patient")


class StagoSegmentSAC(NoRangSegment):
    identifier = HprimString(4)
    carrier = HprimString(11)
    position = HprimString(12)
    additive = Hl7Coding(28)
    dilution = HprimString((30, 4))
    treatment = Hl7Coding(31)
    drug_value = Hl7Coding(42)

    sample_priority_order = HprimForeign("TQ1", "patient")
    orders = HprimForeign("OBR", "patient")
    common_order = HprimForeign("ORC", "patient")
    inventory = HprimForeign("INV", "result")
    specimens = HprimForeign("SPM", "patient")


class StagoSegmentSPM(Segment):
    seq = HprimString(2, default="1")
    specimen_id = HprimString(3)
    specimen_type = Hl7Coding(
        5, default={"value": "PLAS", "description": "Plasma", "coding": "HL70487"}
    )
    role = Hl7Coding(
        12,
        default={"value": "P", "description": "Patient Specimen", "coding": "HL70369"},
    )
    collection_date = HprimDate(18)
    reception_date = HprimDate(19)

    orders = HprimForeign("OBR", "patient")
    infos_samples = HprimForeign("SAC", "patient")
    sample_priority = HprimForeign("TQ1", "patient")


class StagoSegmentINV(Segment):
    substance_id = Hl7Coding(2)
    inventory_container_type = HprimString((4, 1))
    inventory_container_identifier_lot = HprimString((4, 2))
    substance_quantity = HprimDate(9)
    expiration_datetime = HprimDate(13)
    first_used_datetime = HprimDate(14)
    # Lot number reagent, control or calibrator material
    manufacturer_lot_number = HprimString(17)
    onboard_stability_datetime = HprimDate(20)


class StagoSegmentNTE(NoRangSegment):
    comment_source = HprimString(3)
    comment_text = HprimString(4)
    comment_type = Hl7Coding(5)


class StagoSegmentOBR(Segment):
    seq = HprimString(2, default="1")
    placer_order_number = HprimString(3)
    test_identifier = HprimString((5, 1))
    test_text = HprimString((5, 2))
    test_name_coding = HprimString((5, 3), default="99STG")
    action_code = HprimString(12)
    ordering_provider = HprimString((17, 1))  # prescripteur
    ordering_provider_name = HprimString((17, 2))
    ordering_provider_firstname = HprimString((17, 3))

    results = HprimForeign("OBX", "order")
    common_order = HprimForeign("ORC", "order")
    sample_priority = HprimForeign("TQ1", "order")


class StagoSegmentOBX(Segment):
    value_type = HprimString(3)
    observation_id = HprimString((4, 1))
    observation_name = HprimString((4, 2))
    observation_sub_id = HprimString(5)
    value = HprimString(6)
    units = Hl7Coding(7)
    reference_range = HprimString(8)
    flags = HprimList(9, basefield=Hl7Coding(0))
    # F: Final Results
    # X: Results cannot be obtained - Test exception on intrument
    result_status = HprimString(12)
    responsible_observer = HprimString(17)
    equipement_identifier = HprimString((19, 1))
    date_analysis = HprimDate(20)
    observation_type = HprimString(30)

    commentaires = HprimForeign("NTE", "result")
    inventory = HprimForeign("INV", "result")


class StagoSegmentERR(Segment):
    error_location = HprimString(2)
    error_code = HprimString((4, 1))
    severity = HprimString(5)
    application_error = Hl7Coding(6)
    user_message = HprimString(9)


class StagoSegmentEQU(NoRangSegment):
    serial_number = HprimString((2, 3))
    event_date = HprimDate(3)


class StagoSegmentEQP(NoRangSegment):
    event_type = Hl7Coding(2)
    start_date = HprimDate(4)
    end_date = HprimDate(5)


class StagoSegmentORC(Segment):
    order_control = HprimString(2)
    placer_order_number = HprimString(3)
    placer_group_number = HprimString(5)
    order_status = HprimString(6)
    date_transaction = HprimDate(10)
    prescriber_code = HprimString((13, 1))
    prescriber_name = HprimString((13, 2))
    prescriber_firstname = HprimString((13, 3))
    location_facility = HprimString((14, 4))
    location_description = HprimString((14, 7))

    sample_priority = HprimForeign("TQ1", "patient")
    results = HprimForeign("OBX", "order")

    def to_string(self, separators=None):
        self.rang_segment = ""
        return super(StagoSegmentORC, self).to_string(separators=separators)


class StagoSegmentQAK(Segment):
    identification = HprimString(2)
    reponse_status = Hl7Coding(3)
    name = HprimString(4)  # c WOS^Work Order Step^IHELAW No

    orders = HprimForeign("OBR", "patient")
    inventory = HprimForeign("INV", "result")


#  HL70485 - Extended Priority Codes
# S Stat
# R Routine
# A ASAP
class StagoSegmentTQ1(NoRangSegment):  # Value^Description^HL70485
    priority = Hl7Coding(
        10, default={"value": "R", "description": "Routine", "coding": "HL70485"}
    )

    results = HprimForeign("OBX", "order")
    orders = HprimForeign("OBR", "tq1")


class StagoSegmentQPD(Segment):
    query_name = HprimString(2)  # WOS^Work Order
    query_id = HprimString(3)
    sample_id = HprimString(4)

    queries_infos = HprimForeign("RCP", "queries")


class StagoSegmentRCP(Segment):
    query_priority = HprimString(2, default="I")
    response_modality = HprimString(4)  # R^REAL TIME^HL7039


class Driver(AstmMessageV2):
    SPEC = {"code": "sthemo", "label": "StHemo Stago HL7"}
    SEGMENT_CLASS_MAP = {
        "MSH": StagoSegmentMSH,
        "RCP": StagoSegmentRCP,
        "QPD": StagoSegmentQPD,
        "QAK": StagoSegmentQAK,
        "ERR": StagoSegmentERR,
        "PID": StagoSegmentPID,
        "SPM": StagoSegmentSPM,
        "SAC": StagoSegmentSAC,
        "ORC": StagoSegmentORC,
        "OBR": StagoSegmentOBR,
        "OBX": StagoSegmentOBX,
        "INV": StagoSegmentINV,
        "NTE": StagoSegmentNTE,
        "TQ1": StagoSegmentTQ1,
        "EQU": StagoSegmentEQU,
        "PV1": StagoSegmentPV1,
    }
    DEFAULT_SEPARATORS = "|^~\\&"
    FIRST_SEP_INDEX = 3
    SEGMENT_ENDING = "\r"

    OPTIONS = [
        AstmMessageV2.Option(
            "Stago_sender_code", label="Code émeteur", default="Clarisys MCA"
        ),
        AstmMessageV2.Option(
            "Stago_sender_label", label="Libellé émetteur", default="MainFacility"
        ),
        AstmMessageV2.Option(
            "Stago_receiver_code", label="Code récepteur", default="StagoApp"
        ),
        AstmMessageV2.Option(
            "Stago_receiver_label", label="Libellé récepteur", default="LabFacility"
        ),
    ]

    def __init__(self, *args, **kwargs):
        self.sender_code = kwargs.pop("Stago_sender_code", "Clarisys MCA")
        self.sender_label = kwargs.pop("Stago_sender_label", "MainFacility")
        self.receiver_code = kwargs.pop("Stago_receiver_code", "StagoApp")
        self.receiver_label = kwargs.pop("Stago_receiver_label", "LabFacility")
        super(Driver, self).__init__(**kwargs)

    def create(self, *args, **kwargs):
        return StagoSegmentMSH(type_segment="MSH", **kwargs)

    def to_string(self, message, separators=None, **kwargs):
        return super(Driver, self).to_string(message, separators, **kwargs) + "\r"

    @staticmethod
    def get_reagents(seg_inv):
        lots = []
        if seg_inv.inventory_container_type in ("SR", "MR", "DI"):
            reagent_lot = "{}-{}".format(
                seg_inv.manufacturer_lot_number, seg_inv.substance_id.value
            )
            lots.append(
                {
                    "type": "reagent",
                    "lot": reagent_lot,
                    "expiration": seg_inv.expiration_datetime,
                }
            )
        return lots

    def encode_worklist(self, worklist, context):
        msg = self.create(
            sending_application=self.sender_code,
            sending_facility=self.sender_label,
            receiving_application=self.receiver_code,
            receiving_facility=self.receiver_label,
            message_date=context["now"],
            control_id=context.get("control_id", int(time.time() * 1000)),
        )
        for patient in worklist.get_astm_worklist(
            unique_together=["action", "priority", "rerun"],
            priority={"A": "S", "S": "S", worklist.ELSE: "R"},
            action={"del": "CA", worklist.ELSE: "NW"},
        ):
            seg_pid = msg.patients.add(
                patient_id=patient.patient_id,
                lastname=unidecode(patient.last_name),
                firstname=unidecode(patient.first_name),
                birthdate=patient.birthdate,
                sex=patient.sex,
            )
            physician = [
                p for p in patient.physicians if p.type in ("prescriber", "attending")
            ]
            if patient.location:
                seg_pv1 = seg_pid.visits.add(
                    service=patient.services[0].code if patient.services else "",
                    doctor_code=physician[0].code if physician else "",
                    doctor_name=physician[0].last_name if physician else "",
                )
            else:
                seg_pv1 = seg_pid
            for sample, action, priority, rerun in patient.samples:
                seq = 1
                for order in sample["orders"]:
                    seg_spm = seg_pv1.specimens.add(
                        specimen_id=sample.sample_id,
                        collection_date=sample.collection_date,
                    )
                    seg_sac = seg_spm.infos_samples.add(identifier=sample.sample_id)
                    if not rerun and order.dilution:
                        seg_sac.dilution = order.dilution
                    seg_sac.common_order.add(
                        order_control=action, date_transaction=sample.request_date
                    )
                    if priority == "R":
                        priority_label = "Routine"
                    elif priority == "S":
                        priority_label = "STAT"
                    else:
                        priority_label = ""
                    seg_tq1 = seg_sac.sample_priority_order.add(type_segment="TQ1")
                    seg_tq1.priority.value = priority
                    seg_tq1.priority.description = priority_label
                    seg_obr = seg_tq1.orders.add(
                        placer_order_number=context.get(
                            "placer_order_number", int(time.time() * 1000 + seq)
                        ),
                        test_identifier=order.code,
                        test_text=order.code,
                    )
                    for physician in sample.physicians:
                        seg_obr.ordering_provider = physician.code
                    if rerun and order.dilution:
                        seg_obr.infos_tests.add(
                            universal_service_identifier=order.code,
                            test_label=order.code,
                            test_coding_name="99ABT",
                            auto_dilution_type=order.dilution,
                            dilution_coding="99ABT",
                        )
                    seq += 1
        if len(msg.patients):
            return msg
        if context["query"]:
            for qpd in context["query"].queries:
                seg_spm = msg.specimens.add(specimen_id=qpd.sample_id)
                seg_sac = seg_spm.infos_samples.add(
                    query_name=qpd.query_name, identifier=qpd.sample_id
                )
                seg_sac.common_order.add(
                    order_control="DC", date_transaction=context["now"]
                )
            return msg
        return None

    def decode_message(self, message):
        if message.queries:
            worklist = Worklist(context="query")
            for seg_qpd in message.queries:
                sample = worklist.get_or_create(
                    type="patient", sample_id=seg_qpd.sample_id
                )
        else:
            worklist = Worklist(context="result")
            for patient in message.patients:
                if patient.visits:
                    seg_pv1 = patient.visits[0]
                else:
                    seg_pv1 = patient
                for seg_spm in seg_pv1.specimens:
                    is_qc = seg_spm.role.value == "Q"
                    for seg_sac in seg_spm.infos_samples:
                        if seg_sac.carrier and seg_sac.position:
                            rack = "{}-{}".format(seg_sac.carrier, seg_sac.position)
                        else:
                            rack = ""
                        sample = worklist.get_or_create(
                            sample_id=seg_sac.identifier,
                            rack=rack,
                            type="control" if is_qc else "patient",
                        )
                        if is_qc:
                            sample.control_lot = seg_sac.identifier
                        for seg_obr in seg_sac.orders:
                            if seg_obr.common_order:
                                common_order = seg_obr.common_order[0]
                                if seg_obr.common_order[0].sample_priority:
                                    results = common_order.sample_priority[0].results
                                else:
                                    results = seg_obr.results
                                for seg_obx in results:
                                    comments = []
                                    for nte in seg_obx.commentaires:
                                        if not is_qc and nte.comment_source == "Z":
                                            comments.append(
                                                "{} {}".format(
                                                    nte.comment_type.value,
                                                    nte.comment_type.description,
                                                )
                                            )
                                    flags = []
                                    for flag in seg_obx.flags:
                                        if flag.value != '""':
                                            flags.append(flag.value)
                                    dilution = ""
                                    value = seg_obx.value
                                    resu = sample.results.add(
                                        code=seg_obx.observation_id,
                                        value=value,
                                        date=seg_obx.date_analysis,
                                        flags=flags,
                                        instrument=seg_obx.equipement_identifier,
                                        operator=seg_obx.responsible_observer,
                                        unit=seg_obx.units.value,
                                    )
                                    if comments and not is_qc:
                                        resu.comments = comments
                                    if dilution:
                                        resu.dilution = (dilution,)
                                    if not is_qc:
                                        for seg_inv in seg_obx.inventory:
                                            lots = self.get_reagents(seg_inv)
                                            for lot in lots:
                                                resu.lots.add(lot)

        return worklist
